/*   FILE: dmapd.c -- A DMAP server
 * AUTHOR: W. Michael Petullo <mike@flyn.org>
 *   DATE: 01 January 2009 
 *
 * Copyright (c) 2009 W. Michael Petullo <new@flyn.org>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <glib.h>
#include <libdmapsharing/dmap-mdns-publisher.h>
#include <libdmapsharing/dpap-share.h>
#include <libdmapsharing/daap-share.h>

#include "dmapd.h"
#include "dmapd-dmap-container-record.h"
#include "dmapd-dmap-container-db.h"
#include "dmapd-dmap-db.h"
#include "dmapd-dpap-record.h"
#include "dmapd-dpap-record-factory.h"
#include "dmapd-daap-record.h"
#include "dmapd-daap-record-factory.h"

/* For use when deciding whether to serve DAAP or DPAP. */
enum {
	DAAP,
	DPAP
};

GModule *av_meta_module = NULL;
GetGOptionGroupFn dmapd_av_metadata_get_option_group;
AVReadFn dmapd_av_metadata_read;

GModule *photo_meta_module = NULL;
PhotoReadFn dmapd_photo_metadata_read;

GModule *db_module = NULL;
BuildDbStartingAtFn build_db_starting_at;

static GSList *music_dirs = NULL;
static GSList *picture_dirs = NULL;
static gchar *lockpath = NULL;
static gchar *pidpath = NULL;
static gboolean foreground = 0;
static gchar *share_name = NULL;
static gchar *transcode_mimetype;
static gchar *av_module = "gst";
static gchar *photo_module = "graphicsmagick";

static gboolean
add_music_dir (const gchar *option_name,
			const gchar *value,
			gpointer data,
			GError **error)
{
	music_dirs = g_slist_append (music_dirs, g_strdup (value));
}

static gboolean
add_picture_dir (const gchar *option_name,
			  const gchar *value,
			  gpointer data,
			  GError **error)
{
	picture_dirs = g_slist_append (picture_dirs, g_strdup (value));
}

/* FIXME: how to enumerate available transcoding formats? */
static GOptionEntry entries[] = {
	{ "foreground", 'f', 0, G_OPTION_ARG_NONE, &foreground, "Do not fork; remain in foreground", NULL },
	{ "name", 'n', 0, G_OPTION_ARG_STRING, &share_name, "Name of media shares", NULL },
	{ "music-dir", 'm', 0, G_OPTION_ARG_CALLBACK, add_music_dir, "Music directory", NULL },
	{ "picture-dir", 'p', 0, G_OPTION_ARG_CALLBACK, add_picture_dir, "Picture directory", NULL },
	{ "lockpath", 'l', 0, G_OPTION_ARG_STRING, &lockpath, "Path to lockfile", NULL },
	{ "pidpath", 'i', 0, G_OPTION_ARG_STRING, &pidpath, "Path to PID file", NULL },
	{ "transcode-mimetype", 't', 0, G_OPTION_ARG_STRING, &transcode_mimetype, "Target MIME type for real-time transcoding", NULL },
	{ NULL }
};

static char *
default_share_name ()
{
        const gchar *real_name;

        real_name = g_get_real_name ();
        if (strcmp (real_name, "Unknown") == 0) {
                real_name = g_get_user_name ();
        }

        return g_strdup_printf ("%s's Media", real_name);
}

static void
create_share (gint protocol, DMAPDb *db, DMAPContainerDb *container_db)
{
	char *name;

	DMAPShare *share = NULL;

	name = share_name ? share_name : default_share_name ();

	if (protocol == DAAP) {
		g_debug ("Initializing DAAP sharing (%d)", protocol);
		share = DMAP_SHARE (daap_share_new (name, NULL, db, container_db, transcode_mimetype));
	} else {
		g_debug ("Initializing DPAP sharing (%d)", protocol);
		share = DMAP_SHARE (dpap_share_new (name, NULL, db, container_db, transcode_mimetype));
	}

	g_free (name);
}

static void
daemonize (void)
{
	int child, fd;
	char *_lockpath, *pid, *_pidpath;

	if (getppid () == 1)
		return; /* Already a daemon */

	child = fork ();

	if (child < 0)
		g_error ("Error forking");
	
	if (child > 0)
		exit (EXIT_SUCCESS);
	
	setsid ();

	chdir (RUNDIR);

	_lockpath = lockpath ? lockpath : LOCKPATH;
	fd = open (_lockpath, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);

	if (lockf (fd, F_TLOCK, 0) < 0)
		g_error ("Error opening lock file");
	
	_pidpath = pidpath ? pidpath : RUNDIR "/dmapd.pid";
	fd = open (_pidpath, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);

	pid = g_strdup_printf ("%d", getpid ());
	write (fd, pid, strlen (pid));
	g_free (pid);
}

static pid_t
wait_on (pid_t pid, int protocol)
{
	int status;
	waitpid (pid, &status, 0);
	if (WIFEXITED (status)) {
		return WEXITSTATUS (status) ? 1 << protocol : EXIT_SUCCESS;
	} else {
		return EXIT_FAILURE;
	}
}

void load_modules (void)
{
	/* FIXME: the needs to be refactored. */
	gchar *av_meta_module_path;
	gchar *photo_meta_module_path;

	av_meta_module_path = \
		g_strconcat (MODULEDIR, "/libav-meta-", av_module, ".so", NULL);

	av_meta_module = g_module_open (av_meta_module_path, G_MODULE_BIND_LAZY);
	if (! av_meta_module) {
		g_error ("Error opening %s", av_meta_module_path);
	}

	if (! g_module_symbol (av_meta_module, "dmapd_av_metadata_get_option_group", (gpointer *) &dmapd_av_metadata_get_option_group)) {
		g_module_close (av_meta_module);
		g_error ("Symbol missing from module 1");
	}

	if (dmapd_av_metadata_get_option_group == NULL) {
		g_error ("Module function set to NULL 1");
	}

	if (! g_module_symbol (av_meta_module, "dmapd_av_metadata_read", (gpointer *) &dmapd_av_metadata_read)) {
		g_module_close (av_meta_module);
		g_error ("Symbol missing from module 2");
	}

	if (dmapd_av_metadata_read == NULL) {
		g_error ("Module function set to NULL 2");
	}

	photo_meta_module_path = \
		g_strconcat (MODULEDIR, "/libphoto-meta-", photo_module, ".so", NULL);

	photo_meta_module = g_module_open (photo_meta_module_path, G_MODULE_BIND_LAZY);
	if (! photo_meta_module) {
		g_error ("Error opening %s", photo_meta_module_path);
	}

	if (! g_module_symbol (photo_meta_module, "dmapd_photo_metadata_read", (gpointer *) &dmapd_photo_metadata_read)) {
		g_module_close (photo_meta_module);
		g_error ("Symbol missing from module 2");
	}

	if (dmapd_photo_metadata_read == NULL) {
		g_error ("Module function set to NULL 2");
	}

	db_module = g_module_open (MODULEDIR "/libdb-gdir.so", G_MODULE_BIND_LAZY);
	if (! db_module) {
		g_error ("Error opening " MODULEDIR "/libdb-gdir.so");
	}

	if (! g_module_symbol (db_module, "build_db_starting_at", (gpointer *) &build_db_starting_at)) {
		g_module_close (db_module);
		g_error ("Symbol missing from module 3");
	}

	if (build_db_starting_at == NULL) {
		g_error ("Module function set to NULL 3");
	}

	g_free (av_meta_module_path);
	g_free (photo_meta_module_path);
}

static void
debug_printf (const char *log_domain,
	      GLogLevelFlags log_level,
	      const gchar *message,
	      gpointer user_data)
{
	g_printf ("%s\n", message);
}

static void
debug_null (const char *log_domain,
	    GLogLevelFlags log_level,
	    const gchar *message,
	    gpointer user_data)
{
}

int main (int argc, char *argv[])
{
	int exitval = EXIT_SUCCESS;
	pid_t child[2] = { 0, 0 };
	char *root;
	gint protocol;
	GError *error = NULL;
	GOptionContext *context;

	g_type_init ();
	g_thread_init (NULL);

	if (getenv ("DMAPD_DEBUG") != NULL) {
		g_log_set_handler (NULL, G_LOG_LEVEL_DEBUG, debug_printf, NULL);
	} else {
		g_log_set_handler (NULL, G_LOG_LEVEL_DEBUG, debug_null, NULL);
	}

	av_module = getenv ("DMAPD_AV_MODULE");
	av_module = av_module ? av_module : "gst";

	photo_module = getenv ("DMAPD_PHOTO_MODULE");
	photo_module = photo_module ? photo_module : "graphicsmagick";

	load_modules ();

	context = g_option_context_new ("- serve media using DMAP");
	g_option_context_add_main_entries (context, entries, NULL);
	g_option_context_add_group (context, dmapd_av_metadata_get_option_group ());
	if (! g_option_context_parse (context, &argc, &argv, &error)) {
		g_error ("Option parsing failed: %s\n", error->message);
	}

	if (! foreground) {
		daemonize ();
	}

	if (music_dirs) {
#if WITH_DAAP
		/* FIXME: This code is redundant with DPAP code. */
		if ((child[DAAP] = fork ()) == 0) {
			GSList *l;
			DMAPDb *db;
			DMAPRecordFactory *factory;
			DMAPContainerDb *container_db;
			static GMainLoop *loop;

			factory = DMAP_RECORD_FACTORY (dmapd_daap_record_factory_new ());
			db = DMAP_DB (dmapd_dmap_db_new ());
			container_db = DMAP_CONTAINER_DB (dmapd_dmap_container_db_new ());

			for (l = music_dirs; l; l = l->next) {
				build_db_starting_at (l->data, db, container_db, NULL, factory);
			}

			loop = g_main_loop_new (NULL, FALSE);
			create_share (DAAP, DMAP_DB (db), DMAP_CONTAINER_DB (container_db));

			g_main_loop_run (loop);

			exit (EXIT_SUCCESS);
		}
#else
		g_error ("DAAP support not present");
#endif
	}

	if (picture_dirs) {
#ifdef WITH_DPAP
		if ((child[DPAP] = fork ()) == 0) {
			GSList *l;
			DMAPDb *db;
			DMAPRecordFactory *factory;
			DMAPContainerDb *container_db;
			static GMainLoop *loop;

			factory = DMAP_RECORD_FACTORY (dmapd_dpap_record_factory_new ());
			db = DMAP_DB (dmapd_dmap_db_new ());
			container_db = DMAP_CONTAINER_DB (dmapd_dmap_container_db_new ());

			for (l = picture_dirs; l; l = l->next) {
				build_db_starting_at (l->data, db, container_db, NULL, factory);
			}

			loop = g_main_loop_new (NULL, FALSE);
			create_share (DPAP, DMAP_DB (db), DMAP_CONTAINER_DB (container_db));

			g_main_loop_run (loop);

			exit (EXIT_SUCCESS);
		}
#else
		g_error ("DPAP support not present");
#endif
	}

	if (child[DAAP]) {
		exitval |= wait_on (child[DAAP], DAAP);
	}

	if (child[DPAP]) {
		exitval |= wait_on (child[DPAP], DPAP);
	}

	g_module_close (av_meta_module);
	g_module_close (db_module);

	exit(exitval);
}
